@@ -1,7 +1,9 @@
 /*
  * This file is part of bino, a program to play stereoscopic videos.
  *
- * Copyright (C) 2010  Martin Lambers <marlam@marlam.de>
+ * Copyright (C) 2010
+ * Martin Lambers <marlam@marlam.de>
+ * Gabriele Greco <gabrielegreco@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,7 +25,8 @@
 #include <vector>
 #include <stdint.h>
 
-#include <AL/alut.h>
+#include <AL/al.h>
+#include <AL/alc.h>
 
 #include "audio_output.h"
 
@@ -40,19 +43,21 @@ private:
     ALuint _source;
     ALenum _format;
     ALint _state;
-    int _rate;
     int _channels;
+    int _rate;
     int _bits;
     int64_t _basetime;
     int64_t _last_timestamp;
     int64_t _ext_timer_at_last_timestamp;
     int64_t _last_reported_timestamp;
+    ALCdevice *_device;
+    ALCcontext *_context;
 
 public:
     audio_output_openal() throw ();
     ~audio_output_openal();
 
-    virtual void open(int rate, int channels, int bits);
+    virtual void open(int channels, int rate, enum decoder::audio_sample_format format);
 
     virtual int64_t status(size_t *required_data);
     virtual void data(const void *buffer, size_t size);
