@@ -63,6 +63,7 @@ struct internal_stuff
 
     std::vector<int> audio_streams;
     std::vector<AVCodecContext *> audio_codec_ctxs;
+    std::vector<enum decoder::audio_sample_format> audio_sample_formats;
     std::vector<AVCodec *> audio_codecs;
     std::vector<std::deque<AVPacket> > audio_packet_queues;
     std::vector<bool> audio_flush_flags;
@@ -210,7 +211,7 @@ void decoder_ffmpeg::open(const std::string &filename)
                 _stuff->video_codecs[j] = NULL;
                 throw exc(filename + " stream " + str::from(i) + ": cannot open video codec: " + my_av_strerror(e));
             }
-            int bufsize = avpicture_get_size(PIX_FMT_RGB24,
+            int bufsize = avpicture_get_size(PIX_FMT_BGRA,
                     _stuff->video_codec_ctxs[j]->width, _stuff->video_codec_ctxs[j]->height);
             _stuff->packets.push_back(AVPacket());
             _stuff->video_flush_flags.push_back(false);
@@ -222,10 +223,10 @@ void decoder_ffmpeg::open(const std::string &filename)
                 throw exc(HERE + ": " + strerror(ENOMEM));
             }
             avpicture_fill(reinterpret_cast<AVPicture *>(_stuff->out_frames[j]), _stuff->buffers[j],
-                    PIX_FMT_RGB24, _stuff->video_codec_ctxs[j]->width, _stuff->video_codec_ctxs[j]->height);
+                    PIX_FMT_BGRA, _stuff->video_codec_ctxs[j]->width, _stuff->video_codec_ctxs[j]->height);
             _stuff->img_conv_ctxs.push_back(sws_getContext(
                         _stuff->video_codec_ctxs[j]->width, _stuff->video_codec_ctxs[j]->height, _stuff->video_codec_ctxs[j]->pix_fmt,
-                        _stuff->video_codec_ctxs[j]->width, _stuff->video_codec_ctxs[j]->height, PIX_FMT_RGB24,
+                        _stuff->video_codec_ctxs[j]->width, _stuff->video_codec_ctxs[j]->height, PIX_FMT_BGRA,
                         SWS_FAST_BILINEAR, NULL, NULL, NULL));
             if (!_stuff->img_conv_ctxs[j])
             {
@@ -249,6 +250,30 @@ void decoder_ffmpeg::open(const std::string &filename)
                 _stuff->audio_codecs[j] = NULL;
                 throw exc(filename + " stream " + str::from(i) + ": cannot open audio codec: " + my_av_strerror(e));
             }
+            if (_stuff->audio_codec_ctxs[j]->sample_fmt == SAMPLE_FMT_U8)
+            {
+                _stuff->audio_sample_formats.push_back(audio_sample_u8);
+            }
+            else if (_stuff->audio_codec_ctxs[j]->sample_fmt == SAMPLE_FMT_S16)
+            {
+                _stuff->audio_sample_formats.push_back(audio_sample_s16);
+            }
+            else if (_stuff->audio_codec_ctxs[j]->sample_fmt == SAMPLE_FMT_FLT)
+            {
+                _stuff->audio_sample_formats.push_back(audio_sample_f32);
+            }
+            else if (_stuff->audio_codec_ctxs[j]->sample_fmt == SAMPLE_FMT_DBL)
+            {
+                _stuff->audio_sample_formats.push_back(audio_sample_d64);
+            }
+            else
+            {
+                throw exc(filename + " stream " + str::from(i)
+                        + ": cannot handle audio sample format "
+                        + str::from(static_cast<int>(_stuff->audio_codec_ctxs[j]->sample_fmt)));
+                // XXX: Once widely available, use av_get_sample_fmt_name(_stuff->audio_codec_ctxs[j]->sample_fmt)
+                // in this error message.
+            }
             if (_stuff->audio_codec_ctxs[j]->channels < 1
                     || _stuff->audio_codec_ctxs[j]->channels > 8
                     || _stuff->audio_codec_ctxs[j]->channels == 3
@@ -283,8 +308,10 @@ void decoder_ffmpeg::open(const std::string &filename)
     {
         msg::inf("    video stream %d: %dx%d, format %s,",
                 i, video_width(i), video_height(i),
-                _stuff->video_codec_ctxs.at(i)->pix_fmt == PIX_FMT_YUV420P ? "yuv420p"
-                : str::asprintf("%d (converted to rgb24)", _stuff->video_codec_ctxs.at(i)->pix_fmt).c_str());
+                _stuff->video_codec_ctxs.at(i)->pix_fmt == PIX_FMT_YUV420P
+                ? decoder::video_frame_format_name(decoder::frame_format_yuv420p).c_str()
+                : str::asprintf("%d (converted to %s)", _stuff->video_codec_ctxs.at(i)->pix_fmt,
+                    decoder::video_frame_format_name(decoder::frame_format_bgra32).c_str()).c_str());
         msg::inf("        aspect ratio %g:1, %g fps, %g seconds",
                 static_cast<float>(video_aspect_ratio_numerator(i))
                 / static_cast<float>(video_aspect_ratio_denominator(i)),
@@ -294,8 +321,8 @@ void decoder_ffmpeg::open(const std::string &filename)
     }
     for (int i = 0; i < audio_streams(); i++)
     {
-        msg::inf("    audio stream %d: %d channels, %d bits, %d Hz", i,
-                audio_channels(i), audio_bits(i), audio_rate(i));
+        msg::inf("    audio stream %d: %d channels, %d Hz, sample format %s", i,
+                audio_channels(i), audio_rate(i), audio_sample_format_name(audio_sample_format(i)).c_str());
     }
     if (video_streams() == 0 && audio_streams() == 0)
     {
@@ -384,9 +411,11 @@ int64_t decoder_ffmpeg::video_duration(int index) const throw ()
     return duration * 1000000 * time_base.num / time_base.den;
 }
 
-video_frame_format decoder_ffmpeg::video_preferred_frame_format(int index) const throw ()
+enum decoder::video_frame_format decoder_ffmpeg::video_preferred_frame_format(int index) const throw ()
 {
-    return (_stuff->video_codec_ctxs.at(index)->pix_fmt == PIX_FMT_YUV420P ? yuv420p : rgb24);
+    return (_stuff->video_codec_ctxs.at(index)->pix_fmt == PIX_FMT_YUV420P
+            ? decoder::frame_format_yuv420p
+            : decoder::frame_format_bgra32);
 }
 
 int decoder_ffmpeg::audio_rate(int index) const throw ()
@@ -399,9 +428,9 @@ int decoder_ffmpeg::audio_channels(int index) const throw ()
     return _stuff->audio_codec_ctxs.at(index)->channels;
 }
 
-int decoder_ffmpeg::audio_bits(int) const throw ()
+enum decoder::audio_sample_format decoder_ffmpeg::audio_sample_format(int index) const throw ()
 {
-    return 16;  // ffmpeg always decodes to 16 bits
+    return _stuff->audio_sample_formats.at(index);
 }
 
 int64_t decoder_ffmpeg::audio_duration(int index) const throw ()
@@ -519,9 +548,9 @@ void decoder_ffmpeg::get_video_frame(int video_stream, video_frame_format fmt,
     line_size[0] = 0;
     line_size[1] = 0;
     line_size[2] = 0;
-    if (fmt == yuv420p)
+    if (fmt == decoder::frame_format_yuv420p)
     {
-        if (video_preferred_frame_format(video_stream) == yuv420p)
+        if (video_preferred_frame_format(video_stream) == decoder::frame_format_yuv420p)
         {
             data[0] = _stuff->frames[video_stream]->data[0];
             data[1] = _stuff->frames[video_stream]->data[1];
@@ -573,7 +602,7 @@ int64_t decoder_ffmpeg::read_audio_data(int audio_stream, void *buffer, size_t s
             _stuff->audio_buffers[audio_stream].erase(
                     _stuff->audio_buffers[audio_stream].begin(),
                     _stuff->audio_buffers[audio_stream].begin() + remaining);
-            buffer = reinterpret_cast<int16_t *>(reinterpret_cast<unsigned char *>(buffer) + remaining);
+            buffer = reinterpret_cast<unsigned char *>(buffer) + remaining;
             i += remaining;
         }
         if (_stuff->audio_buffers[audio_stream].size() == 0)
@@ -591,7 +620,7 @@ int64_t decoder_ffmpeg::read_audio_data(int audio_stream, void *buffer, size_t s
             _stuff->audio_packet_queues[audio_stream].pop_front();
             if (timestamp == std::numeric_limits<int64_t>::min())
             {
-                timestamp = _stuff->packets[audio_stream].dts * 1000000
+                timestamp = packet.dts * 1000000
                     * _stuff->format_ctx->streams[_stuff->audio_streams[audio_stream]]->time_base.num 
                     / _stuff->format_ctx->streams[_stuff->audio_streams[audio_stream]]->time_base.den;
             }
@@ -601,7 +630,7 @@ int64_t decoder_ffmpeg::read_audio_data(int audio_stream, void *buffer, size_t s
             while (tmppacket.size > 0)
             {
                 int tmpbuf_size = (AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2;
-                unsigned char tmpbuf[tmpbuf_size];
+                unsigned char *tmpbuf = new unsigned char[tmpbuf_size];
                 int len = avcodec_decode_audio3(_stuff->audio_codec_ctxs[audio_stream],
                         reinterpret_cast<int16_t *>(tmpbuf), &tmpbuf_size, &tmppacket);
                 if (len < 0)
@@ -616,8 +645,10 @@ int64_t decoder_ffmpeg::read_audio_data(int audio_stream, void *buffer, size_t s
                     continue;
                 }
                 // Put it in the decoded audio data buffer
-                _stuff->audio_buffers[audio_stream].resize(tmpbuf_size);
-                memcpy(&(_stuff->audio_buffers[audio_stream][0]), tmpbuf, tmpbuf_size);
+                size_t old_size = _stuff->audio_buffers[audio_stream].size();
+                _stuff->audio_buffers[audio_stream].resize(old_size + tmpbuf_size);
+                memcpy(&(_stuff->audio_buffers[audio_stream][old_size]), tmpbuf, tmpbuf_size);
+                delete[] tmpbuf;
             }
             
             av_free_packet(&packet);
