@@ -12,6 +12,25 @@
 #include "player_equalizer.h"
 
 
+/*
+ * Every eq::Node has a special player: player_eq_node.
+ * These node players do not control the video playing themselves. Instead,
+ * they are told what to do via the frame data.
+ *
+ * eq::Config has the master player object. This master also plays the sound,
+ * to be able to synchronize the video to it.
+ *
+ * The application node reuses the player of eq::Config, so that it does not
+ * need to run two players.
+ *
+ * Each eq::Window has a special video_output_opengl: video_output_opengl_eq_window.
+ * It manages the video textures.
+ *
+ * Each eq::Channel than calls the window's display function to render its subset of
+ * the video.
+ */
+
+
 /*
  * player_eq_node
  *
@@ -59,11 +78,21 @@ public:
         return true;
     }
 
+    video_output_state &eq_video_state()
+    {
+        return video_state();
+    }
+
+    void eq_seek_to(int64_t seek_to)
+    {
+        get_input()->seek(seek_to);
+    }
+
     void eq_read_frame()        // Only called on slave nodes to force reading the next frame
     {
         if (get_input()->read_video_frame() < 0)
         {
-            msg::wrn("reading input frame failed (EOF?)");
+            msg::err("reading input frame failed (EOF?)");
             abort();
         }
     }
@@ -73,9 +102,9 @@ public:
         prepare_video_frame(vo);
     }
 
-    void eq_run_step(bool *more_steps, bool *prep_frame, bool *drop_frame, bool *display_frame)
+    void eq_run_step(bool *more_steps, int64_t *seek_to, bool *prep_frame, bool *drop_frame, bool *display_frame)
     {
-        run_step(more_steps, prep_frame, drop_frame, display_frame);
+        run_step(more_steps, seek_to, prep_frame, drop_frame, display_frame);
     }
 
     void eq_get_frame()
@@ -98,21 +127,20 @@ public:
 class video_output_opengl_eq_window : public video_output_opengl
 {
 public:
-    video_output_opengl_eq_window() : video_output_opengl()
+    video_output_opengl_eq_window() : video_output_opengl(false)
     {
     }
 
     void eq_display(enum video_output::mode mode, float x, float y, float w, float h)
     {
-        // TODO: x, y, w, h describe the quad, set by the channel
-        video_output_opengl::display(mode);
+        video_output_opengl::display(mode, x, y, w, h);
     }
 
     void eq_initialize(int src_width, int src_height, float src_aspect_ratio,
             enum decoder::video_frame_format src_preferred_frame_format,
             bool have_pixel_buffer_object, bool have_texture_non_power_of_two, bool have_fragment_shader)
     {
-        set_mode(stereo);       // just to ensure that prepare() does the right thing
+        set_mode(stereo);       // just to ensure that prepare() handles both left and right view
         set_source_info(src_width, src_height, src_aspect_ratio, src_preferred_frame_format);
         initialize(have_pixel_buffer_object, have_texture_non_power_of_two, have_fragment_shader);
     }
@@ -127,7 +155,12 @@ public:
         swap_tex_set();
     }
 
-    // The rest of the interface is irrelevant for Equalizer
+    void eq_set_state(const video_output_state &video_state)
+    {
+        set_state(video_state);
+    }
+
+    // The rest of the video_output_opengl interface is irrelevant for Equalizer
     virtual int window_pos_x() { return 0; }
     virtual int window_pos_y() { return 0; }
     virtual void receive_notification(const notification &) {}
@@ -147,9 +180,14 @@ class eq_init_data : public eq::net::Object
 public:
     uint32_t frame_data_id;
     player_init_data init_data;
+    struct { float x, y, w, h; } canvas_video_area;
 
     eq_init_data() : frame_data_id(EQ_ID_INVALID), init_data()
     {
+        canvas_video_area.x = 0.0f;
+        canvas_video_area.y = 0.0f;
+        canvas_video_area.w = 1.0f;
+        canvas_video_area.h = 1.0f;
     }
 
     ~eq_init_data()
@@ -178,6 +216,10 @@ protected:
         s11n::save(oss, init_data.video_state.fullscreen);
         s11n::save(oss, init_data.video_state.swap_eyes);
         s11n::save(oss, init_data.video_flags);
+        s11n::save(oss, canvas_video_area.x);
+        s11n::save(oss, canvas_video_area.y);
+        s11n::save(oss, canvas_video_area.w);
+        s11n::save(oss, canvas_video_area.h);
         os << oss.str();
     }
 
@@ -202,6 +244,10 @@ protected:
         s11n::load(iss, init_data.video_state.fullscreen);
         s11n::load(iss, init_data.video_state.swap_eyes);
         s11n::load(iss, init_data.video_flags);
+        s11n::load(iss, canvas_video_area.x);
+        s11n::load(iss, canvas_video_area.y);
+        s11n::load(iss, canvas_video_area.w);
+        s11n::load(iss, canvas_video_area.h);
     }
 };
 
@@ -213,14 +259,14 @@ class eq_frame_data : public eq::net::Object
 {
 public:
     video_output_state video_state;
-    int64_t seek_request;
+    int64_t seek_to;
     bool prep_frame;
     bool drop_frame;
     bool display_frame;
 
 public:
     eq_frame_data()
-        : video_state(), seek_request(0),
+        : video_state(), seek_to(-1),
         prep_frame(false), drop_frame(false), display_frame(false)
     {
     }
@@ -240,7 +286,7 @@ protected:
         s11n::save(oss, video_state.saturation);
         s11n::save(oss, video_state.fullscreen);
         s11n::save(oss, video_state.swap_eyes);
-        s11n::save(oss, seek_request);
+        s11n::save(oss, seek_to);
         s11n::save(oss, prep_frame);
         s11n::save(oss, drop_frame);
         s11n::save(oss, display_frame);
@@ -258,7 +304,7 @@ protected:
         s11n::load(iss, video_state.saturation);
         s11n::load(iss, video_state.fullscreen);
         s11n::load(iss, video_state.swap_eyes);
-        s11n::load(iss, seek_request);
+        s11n::load(iss, seek_to);
         s11n::load(iss, prep_frame);
         s11n::load(iss, drop_frame);
         s11n::load(iss, display_frame);
@@ -273,18 +319,21 @@ class eq_config : public eq::Config
 {
 private:
     bool _is_master_config;
-    eq_init_data _eq_init_data;               // Master eq_init_data instance
-    eq_frame_data _eq_frame_data;             // Master eq_frame_data instance
-    player_eq_node _player;
+    eq_init_data _eq_init_data;         // Master eq_init_data instance
+    eq_frame_data _eq_frame_data;       // Master eq_frame_data instance
+    player_eq_node _player;             // Master player
+    controller _controller;             // Sends commands to the player
 
 public:
+    // Source video properties:
     int src_width, src_height;
     float src_aspect_ratio;
     enum decoder::video_frame_format src_preferred_frame_format;
 
 public:
     eq_config(eq::ServerPtr parent)
-        : eq::Config(parent), _is_master_config(false), _eq_init_data(), _eq_frame_data(), _player(),
+        : eq::Config(parent), _is_master_config(false), _eq_init_data(), _eq_frame_data(),
+        _player(), _controller(false),
         src_width(-1), src_height(-1), src_aspect_ratio(0.0f),
         src_preferred_frame_format(decoder::frame_format_yuv420p)
     {
@@ -301,28 +350,51 @@ public:
         msg::dbg(HERE);
         // If this function is called, then this is the master config
         _is_master_config = true;
-        // Initialize master instances
+        // Initialize master init/frame data instances
         _eq_init_data.init_data = init_data;
         _eq_frame_data.video_state = _eq_init_data.init_data.video_state;
-        // Register master instances
-        registerObject(&_eq_frame_data);
-        _eq_init_data.frame_data_id = _eq_frame_data.getID();
-        registerObject(&_eq_init_data);
-        // Initialize player
+        // Initialize master player
         _player.eq_make_master();
         if (!_player.eq_init(init_data, &src_width, &src_height, &src_aspect_ratio, &src_preferred_frame_format))
         {
             return false;
         }
+        // Find region of canvas to use, depending on the video aspect ratio
+        if (getCanvases().size() < 1)
+        {
+            msg::err("no canvas in Equalizer configuration");
+            abort();
+        }
+        float canvas_w = getCanvases()[0]->getWall().getWidth();
+        float canvas_h = getCanvases()[0]->getWall().getHeight();
+        float canvas_aspect_ratio = canvas_w / canvas_h;
+        _eq_init_data.canvas_video_area.w = 1.0f;
+        _eq_init_data.canvas_video_area.h = 1.0f;
+        if (src_aspect_ratio > canvas_aspect_ratio)
+        {
+            // need black borders top and bottom
+            _eq_init_data.canvas_video_area.h = canvas_aspect_ratio / src_aspect_ratio;
+        }
+        else
+        {
+            // need black borders left and right
+            _eq_init_data.canvas_video_area.w = src_aspect_ratio / canvas_aspect_ratio;
+        }
+        _eq_init_data.canvas_video_area.x = (1.0f - _eq_init_data.canvas_video_area.w) / 2.0f;
+        _eq_init_data.canvas_video_area.y = (1.0f - _eq_init_data.canvas_video_area.h) / 2.0f;
+        msg::inf("equalizer canvas:");
+        msg::inf("    %gx%g, aspect ratio %g:1", canvas_w, canvas_h, canvas_w / canvas_h);
+        msg::inf("    area for %g:1 video: [ %g %g %g %g ]", src_aspect_ratio,
+                _eq_init_data.canvas_video_area.x, _eq_init_data.canvas_video_area.y,
+                _eq_init_data.canvas_video_area.w, _eq_init_data.canvas_video_area.h);
+        // Register master instances
+        registerObject(&_eq_frame_data);
+        _eq_init_data.frame_data_id = _eq_frame_data.getID();
+        registerObject(&_eq_init_data);
         msg::dbg(HERE);
         return eq::Config::init(_eq_init_data.getID());
     }
 
-    void prep_frame(video_output *vo)
-    {
-        _player.eq_prep_frame(vo);
-    }
-
     virtual bool exit()
     {
         msg::dbg(HERE);
@@ -331,52 +403,108 @@ public:
         deregisterObject(&_eq_init_data);
         deregisterObject(&_eq_frame_data);
         _eq_init_data.frame_data_id = EQ_ID_INVALID;
+        // Cleanup
+        _player.close();
         msg::dbg(HERE);
         return ret;
     }
 
     virtual uint32_t startFrame()
     {
-        msg::dbg(HERE);
-        bool more_steps, prep_frame, drop_frame, display_frame;
-        _player.eq_run_step(&more_steps, &prep_frame, &drop_frame, &display_frame);
+        // Run one player step to find out what to do
+        bool more_steps;
+        _player.eq_run_step(&more_steps, &_eq_frame_data.seek_to,
+                &_eq_frame_data.prep_frame, &_eq_frame_data.drop_frame, &_eq_frame_data.display_frame);
         if (!more_steps)
         {
             this->exit();
         }
-        if (prep_frame)
-        {
-            _player.eq_get_frame();
-        }
-        else if (drop_frame)
-        {
-            _player.eq_release_frame();
-        }
-        _eq_frame_data.prep_frame = prep_frame;
-        _eq_frame_data.drop_frame = drop_frame;
-        _eq_frame_data.display_frame = display_frame;
-
-        // Update frame data
+        // Update the video state for all (it might have changed via handleEvent())
+        _eq_frame_data.video_state = _player.eq_video_state();
+        // Commit the updated frame data
         const uint32_t version = _eq_frame_data.commit();
         // Start this frame with the committed frame data
-        msg::dbg(HERE);
         return eq::Config::startFrame(version);
     }
 
-    virtual bool handleEvent(const eq::ConfigEvent *event);
-};
-
-bool eq_config::handleEvent(const eq::ConfigEvent *event)
-{
-    if (eq::Config::handleEvent(event))
+    virtual bool handleEvent(const eq::ConfigEvent *event)
     {
+        if (eq::Config::handleEvent(event))
+        {
+            return true;
+        }
+        if (event->data.type == eq::Event::KEY_PRESS)
+        {
+            switch (event->data.keyPress.key)
+            {
+            case 'q':
+                _controller.send_cmd(command::toggle_play);
+                break;
+            case 's':
+                _controller.send_cmd(command::toggle_swap_eyes);
+                break;
+            case 'f':
+                /* fullscreen toggling not supported with Equalizer */
+                break;
+            case 'c':
+                /* window centering not supported with Equalizer */
+                break;
+            case ' ':
+            case 'p':
+                _controller.send_cmd(command::toggle_pause);
+                break;
+            case '1':
+                _controller.send_cmd(command::adjust_contrast, -0.05f);
+                break;
+            case '2':
+                _controller.send_cmd(command::adjust_contrast, +0.05f);
+                break;
+            case '3':
+                _controller.send_cmd(command::adjust_brightness, -0.05f);
+                break;
+            case '4':
+                _controller.send_cmd(command::adjust_brightness, +0.05f);
+                break;
+            case '5':
+                _controller.send_cmd(command::adjust_hue, -0.05f);
+                break;
+            case '6':
+                _controller.send_cmd(command::adjust_hue, +0.05f);
+                break;
+            case '7':
+                _controller.send_cmd(command::adjust_saturation, -0.05f);
+                break;
+            case '8':
+                _controller.send_cmd(command::adjust_saturation, +0.05f);
+                break;
+            case eq::KC_LEFT:
+                _controller.send_cmd(command::seek, -10.0f);
+                break;
+            case eq::KC_RIGHT:
+                _controller.send_cmd(command::seek, +10.0f);
+                break;
+            case eq::KC_UP:
+                _controller.send_cmd(command::seek, -60.0f);
+                break;
+            case eq::KC_DOWN:
+                _controller.send_cmd(command::seek, +60.0f);
+                break;
+            case eq::KC_PAGE_UP:
+                _controller.send_cmd(command::seek, -600.0f);
+                break;
+            case eq::KC_PAGE_DOWN:
+                _controller.send_cmd(command::seek, +600.0f);
+                break;
+            }
+        }
         return true;
     }
 
-    /* ... */
-
-    return false;
-}
+    player_eq_node *player()
+    {
+        return &_player;
+    }
+};
 
 /*
  * eq_node
@@ -427,9 +555,9 @@ protected:
             return false;
         }
 
-        // Create decoders and input
         msg::set_level(init_data.init_data.log_level);
         msg::dbg(HERE);
+        // Create decoders and input
         if (!_is_app_node)
         {
             if (!_player.eq_init(init_data.init_data, &src_width, &src_height, &src_aspect_ratio, &src_preferred_frame_format))
@@ -456,37 +584,61 @@ protected:
         config->unmapObject(&frame_data);
         // Unmap our InitData instance
         config->unmapObject(&init_data);
+        // Cleanup
+        _player.close();
         msg::dbg(HERE);
         return eq::Node::configExit();
     }
 
     virtual void frameStart(const uint32_t frameID, const uint32_t frameNumber)
     {
-        msg::dbg(HERE);
-        // Update our FrameData instance from the master instance
+        // Update our frame data
         frame_data.sync(frameID);
-
-        if (!_is_app_node)
+        // Do as we're told
+        if (_is_app_node)
         {
+            eq_config *config = static_cast<eq_config *>(getConfig());
+            if (frame_data.prep_frame)
+            {
+                config->player()->eq_get_frame();
+            }
+            if (frame_data.drop_frame)
+            {
+                config->player()->eq_release_frame();
+            }
+        }
+        else
+        {
+            if (frame_data.seek_to >= 0)
+            {
+                _player.eq_seek_to(frame_data.seek_to);
+            }
             if (frame_data.prep_frame)
             {
                 _player.eq_read_frame();
                 _player.eq_get_frame();
             }
-            else if (frame_data.drop_frame)
+            if (frame_data.drop_frame)
             {
                 _player.eq_read_frame();
                 _player.eq_release_frame();
             }
         }
         startFrame(frameNumber);
-        msg::dbg(HERE);
     }
 
     virtual void frameFinish(const uint32_t, const uint32_t frameNumber)
     {
-        msg::dbg(HERE);
-        if (!_is_app_node)
+        // Do as we're told
+        if (_is_app_node)
+        {
+            eq_config *config = static_cast<eq_config *>(getConfig());
+            if (frame_data.prep_frame)
+            {
+                config->player()->eq_release_frame();
+            }
+        }
+        else
         {
             if (frame_data.prep_frame)
             {
@@ -494,29 +646,25 @@ protected:
             }
         }
         releaseFrame(frameNumber);
-        msg::dbg(HERE);
     }
 
 public:
     void prep_frame(video_output *vo, enum decoder::video_frame_format fmt)
     {
-        msg::dbg(HERE);
         if (fmt != src_preferred_frame_format)
         {
             msg::err("cannot provide video in requested frame format");
             abort();
         }
-        msg::dbg(HERE);
         if (_is_app_node)
         {
             eq_config *config = static_cast<eq_config *>(getConfig());
-            config->prep_frame(vo);
+            config->player()->eq_prep_frame(vo);
         }
         else
         {
             _player.eq_prep_frame(vo);
         }
-        msg::dbg(HERE);
     }
 };
 
@@ -546,9 +694,9 @@ public:
     {
     }
 
-    void display(enum video_output::mode mode)
+    void display(enum video_output::mode mode, float x, float y, float w, float h)
     {
-        _video_output.eq_display(mode, -1.0f, -1.0f, 1.0f, 1.0f);
+        _video_output.eq_display(mode, x, y, w, h);
     }
 
 protected:
@@ -589,27 +737,26 @@ protected:
 
     virtual void frameStart(const uint32_t, const uint32_t frameNumber)
     {
-        msg::dbg(HERE);
+        // Get frame data via from the node
         eq_node *node = static_cast<eq_node *>(getNode());
+        _video_output.eq_set_state(node->frame_data.video_state);
+        // Do as we're told
         if (node->frame_data.prep_frame)
         {
-            makeCurrent();
+            makeCurrent();      // XXX Is this necessary?
             node->prep_frame(&_video_output, _video_output.frame_format());
         }
-        else if (node->frame_data.display_frame)
+        if (node->frame_data.display_frame)
         {
-            makeCurrent();
+            makeCurrent();      // XXX Is this necessary?
             _video_output.eq_swap_tex_set();
         }
         startFrame(frameNumber);
-        msg::dbg(HERE);
     }
 
     virtual void frameFinish(const uint32_t, const uint32_t frameNumber)
     {
-        msg::dbg(HERE);
         releaseFrame(frameNumber);
-        msg::dbg(HERE);
     }
 };
 
@@ -629,26 +776,33 @@ public:
 protected:
     virtual void frameDraw(const uint32_t frameID)
     {
-        msg::dbg(HERE);
         // Let Equalizer initialize some stuff
         eq::Channel::frameDraw(frameID);
-
-        // Get the current state
-        //eq_node *node = static_cast<eq_node *>(getNode());
-        eq_window *window = static_cast<eq_window *>(getWindow());
-
-        eq::PixelViewport eq_viewport = getPixelViewport();
-        eq::Frustumf eq_frustum = getFrustum();
-        eq::Matrix4f eq_head_transform = getHeadTransform();
-
+        // Get the canvas video area and the canvas channel area
+        eq_node *node = static_cast<eq_node *>(getNode());
+        const struct { float x, y, w, h; } canvas_video_area =
+        {
+            node->init_data.canvas_video_area.x,
+            node->init_data.canvas_video_area.y,
+            node->init_data.canvas_video_area.w,
+            node->init_data.canvas_video_area.h
+        };
+        const eq::Viewport &canvas_channel_area = getViewport();
+        // Determine the video quad to render
+        float quad_x = ((canvas_video_area.x - canvas_channel_area.x) / canvas_channel_area.w - 0.5) * 2.0f;
+        float quad_y = ((canvas_video_area.y - canvas_channel_area.y) / canvas_channel_area.h - 0.5) * 2.0f;
+        float quad_w = 2.0f * canvas_video_area.w / canvas_channel_area.w;
+        float quad_h = 2.0f * canvas_video_area.h / canvas_channel_area.h;
+        // Display
         glEnable(GL_TEXTURE_2D);
         glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
         glMatrixMode(GL_PROJECTION);
         glLoadIdentity();
         glMatrixMode(GL_MODELVIEW);
         glLoadIdentity();
-        window->display(getEye() == eq::EYE_RIGHT ? video_output::mono_right : video_output::mono_left);
-        msg::dbg(HERE);
+        eq_window *window = static_cast<eq_window *>(getWindow());
+        window->display(getEye() == eq::EYE_RIGHT ? video_output::mono_right : video_output::mono_left,
+                quad_x, quad_y, quad_w, quad_h);
     }
 };
 
@@ -696,16 +850,15 @@ player_equalizer::player_equalizer(int *argc, char *argv[])
     _node_factory = static_cast<void *>(new eq_node_factory);
     if (!eq::init(*argc, argv, static_cast<eq::NodeFactory *>(_node_factory)))
     {
-        throw exc("Equalizer initialization failed");
+        throw exc("equalizer initialization failed");
     }
-
     /* Get a configuration */
     _config = static_cast<void *>(eq::getConfig(*argc, argv));
     // The following code is only executed on the application node because
     // eq::getConfig() does not return on other nodes.
     if (!_config)
     {
-        throw exc("Cannot get Equalizer configuration");
+        throw exc("cannot get equalizer configuration");
     }
 }
 
@@ -719,7 +872,7 @@ void player_equalizer::open(const player_init_data &init_data)
     eq_config *config = static_cast<eq_config *>(_config);
     if (!config->init(init_data))
     {
-        throw exc("Equalizer configuration initialization failed");
+        throw exc("equalizer configuration initialization failed");
     }
 }
 
