@@ -19,6 +19,7 @@
 
 #include "config.h"
 
+#include <QCoreApplication>
 #include <QMainWindow>
 #include <QCloseEvent>
 #include <QMenu>
@@ -32,6 +33,7 @@
 #include <QPushButton>
 #include <QComboBox>
 #include <QTimer>
+#include <QFile>
 
 #include "player_qt.h"
 #include "qt_app.h"
@@ -54,7 +56,8 @@ void player_qt_internal::open(const player_init_data &init_data)
     create_input(init_data.input_mode);
     create_audio_output();
     set_video_output(_vo);
-    open_video_output(init_data.video_mode, init_data.video_state, init_data.video_flags);
+    video_state() = init_data.video_state;
+    open_video_output(init_data.video_mode, init_data.video_flags);
 }
 
 void player_qt_internal::close()
@@ -81,17 +84,18 @@ void player_qt_internal::receive_notification(const notification &note)
     if (note.type == notification::play)
     {
         _playing = note.current.flag;
-    }    
+    }
 }
 
 bool player_qt_internal::playloop_step()
 {
     bool more_steps;
+    int64_t seek_to;
     bool prep_frame;
     bool drop_frame;
     bool display_frame;
 
-    run_step(&more_steps, &prep_frame, &drop_frame, &display_frame);
+    run_step(&more_steps, &seek_to, &prep_frame, &drop_frame, &display_frame);
     if (!more_steps)
     {
         return false;
@@ -102,11 +106,11 @@ bool player_qt_internal::playloop_step()
         prepare_video_frame(get_video_output());
         release_video_frame();
     }
-    else if (drop_frame)
+    if (drop_frame)
     {
         get_input()->release_video_frame();
     }
-    else if (display_frame)
+    if (display_frame)
     {
         get_video_output()->activate();
     }
@@ -125,7 +129,8 @@ QWidget *player_qt_internal::video_output_widget()
 }
 
 
-in_out_widget::in_out_widget(QWidget *parent) : QWidget(parent), _lock(false)
+in_out_widget::in_out_widget(QSettings *settings, QWidget *parent)
+    : QWidget(parent), _settings(settings), _lock(false)
 {
     QGridLayout *layout0 = new QGridLayout;
     QLabel *input_label = new QLabel("Input:");
@@ -138,6 +143,7 @@ in_out_widget::in_out_widget(QWidget *parent) : QWidget(parent), _lock(false)
     _input_combobox->addItem("Left/right");
     _input_combobox->addItem("Left/right, half width");
     _input_combobox->addItem("Even/odd rows");
+    connect(_input_combobox, SIGNAL(currentIndexChanged(int)), this, SLOT(input_changed()));
     layout0->addWidget(_input_combobox, 0, 1);
     QLabel *output_label = new QLabel("Output:");
     layout0->addWidget(output_label, 1, 0);
@@ -186,27 +192,9 @@ in_out_widget::~in_out_widget()
 {
 }
 
-void in_out_widget::swap_eyes_changed()
+void in_out_widget::set_input(enum input::mode m)
 {
-    if (!_lock)
-    {
-        send_cmd(command::toggle_swap_eyes);
-    }
-}
-
-void in_out_widget::fullscreen_pressed()
-{
-    send_cmd(command::toggle_fullscreen);
-}
-
-void in_out_widget::center_pressed()
-{
-    send_cmd(command::center);
-}
-
-void in_out_widget::update(const player_init_data &init_data, bool have_valid_input, bool playing)
-{
-    switch (init_data.input_mode)
+    switch (m)
     {
     default:
     case input::mono:
@@ -231,7 +219,11 @@ void in_out_widget::update(const player_init_data &init_data, bool have_valid_in
         _input_combobox->setCurrentIndex(6);
         break;
     }
-    switch (init_data.video_mode)
+}
+
+void in_out_widget::set_output(enum video_output::mode m)
+{
+    switch (m)
     {
     default:
     case video_output::mono_left:
@@ -277,6 +269,46 @@ void in_out_widget::update(const player_init_data &init_data, bool have_valid_in
         _output_combobox->setCurrentIndex(13);
         break;
     }
+}
+
+void in_out_widget::input_changed()
+{
+    _settings->beginGroup("Session");
+    if (_input_combobox->currentIndex() == 0)
+    {
+        set_output(static_cast<enum video_output::mode>(
+                    _settings->value("2d-output-mode", static_cast<int>(video_output::mono_left)).toInt()));
+    }
+    else
+    {
+        set_output(static_cast<enum video_output::mode>(
+                    _settings->value("3d-output-mode", static_cast<int>(video_output::anaglyph_red_cyan_dubois)).toInt()));
+    }
+    _settings->endGroup();
+}
+
+void in_out_widget::swap_eyes_changed()
+{
+    if (!_lock)
+    {
+        send_cmd(command::toggle_swap_eyes);
+    }
+}
+
+void in_out_widget::fullscreen_pressed()
+{
+    send_cmd(command::toggle_fullscreen);
+}
+
+void in_out_widget::center_pressed()
+{
+    send_cmd(command::center);
+}
+
+void in_out_widget::update(const player_init_data &init_data, bool have_valid_input, bool playing)
+{
+    set_input(init_data.input_mode);
+    set_output(init_data.video_mode);
     _lock = true;
     _swap_eyes_button->setChecked(init_data.video_state.swap_eyes);
     _lock = false;
@@ -374,7 +406,8 @@ void in_out_widget::receive_notification(const notification &note)
 }
 
 
-controls_widget::controls_widget(QWidget *parent) : QWidget(parent), _playing(false)
+controls_widget::controls_widget(QSettings *settings, QWidget *parent)
+    : QWidget(parent), _settings(settings), _playing(false)
 {
     QGridLayout *layout = new QGridLayout;
     _play_button = new QPushButton(QIcon(":icons/play.png"), "");
@@ -523,8 +556,8 @@ void controls_widget::receive_notification(const notification &note)
 }
 
 
-main_window::main_window(const player_init_data &init_data)
-    : _last_open_dir(QDir::current()), _player(NULL), _init_data(init_data), _stop_request(false)
+main_window::main_window(QSettings *settings, const player_init_data &init_data)
+    : _settings(settings), _player(NULL), _init_data(init_data), _stop_request(false)
 {
     // Application properties
     setWindowTitle(PACKAGE_NAME);
@@ -537,9 +570,9 @@ main_window::main_window(const player_init_data &init_data)
     _video_output->open(decoder::frame_format_bgra32, 1, 1, 1.0f, video_output::mono_left, video_output_state(), 0, 0, 0);
     _video_widget = _video_output->widget();
     _layout->addWidget(_video_widget, 0, 0);
-    _in_out_widget = new in_out_widget(central_widget);
+    _in_out_widget = new in_out_widget(_settings, central_widget);
     _layout->addWidget(_in_out_widget, 1, 0);
-    _controls_widget = new controls_widget(central_widget);
+    _controls_widget = new controls_widget(_settings, central_widget);
     _layout->addWidget(_controls_widget, 2, 0);
     _layout->setRowStretch(0, 1);
     _layout->setColumnStretch(0, 1);
@@ -586,7 +619,7 @@ main_window::main_window(const player_init_data &init_data)
         QStringList filenames;
         for (size_t i = 0; i < init_data.filenames.size(); i++)
         {
-            filenames.push_back(init_data.filenames[i].c_str());
+            filenames.push_back(QFile::decodeName(init_data.filenames[i].c_str()));
         }
         open(filenames, false);
     }
@@ -633,6 +666,23 @@ void main_window::receive_notification(const notification &note)
             {
                 _stop_request = true;
             }
+            _settings->beginGroup("Video");
+            if (_init_data.filenames.size() == 1)
+            {
+                _settings->setValue(QFileInfo(QFile::decodeName(_init_data.filenames[0].c_str())).fileName(),
+                        QString(input::mode_name(_init_data.input_mode).c_str()));
+            }
+            _settings->endGroup();
+            _settings->beginGroup("Session");
+            if (_init_data.input_mode == input::mono)
+            {
+                _settings->setValue("2d-output-mode", QString(video_output::mode_name(_init_data.video_mode).c_str()));
+            }
+            else
+            {
+                _settings->setValue("3d-output-mode", QString(video_output::mode_name(_init_data.video_mode).c_str()));
+            }
+            _settings->endGroup();
             _in_out_widget->update(_init_data, true, true);
             _controls_widget->update(_init_data, true, true);
             _video_widget->setFocus(Qt::OtherFocusReason);
@@ -690,12 +740,36 @@ void main_window::open(QStringList filenames, bool automatic)
     if (automatic)
     {
         _init_data.input_mode = input::automatic;
-        _init_data.video_mode = video_output::automatic;
     }
     if (open_player())
     {
-        _init_data.input_mode = _player->input_mode();
-        _init_data.video_mode = _player->video_mode();
+        _settings->beginGroup("Video");
+        if (_init_data.filenames.size() == 1)
+        {
+            QString fallback_mode_name = QString(input::mode_name(_player->input_mode()).c_str());
+            QString filename = QFileInfo(QFile::decodeName(_init_data.filenames[0].c_str())).fileName();
+            QString mode_name = _settings->value(filename, fallback_mode_name).toString();
+            _init_data.input_mode = input::mode_from_name(mode_name.toStdString());
+        }
+        else
+        {
+            _init_data.input_mode = _player->input_mode();
+        }
+        _settings->endGroup();
+        _settings->beginGroup("Session");
+        if (_init_data.input_mode == input::mono)
+        {
+            QString fallback_mode_name = QString(video_output::mode_name(video_output::mono_left).c_str());
+            QString mode_name = _settings->value(QString("2d-output-mode"), fallback_mode_name).toString();
+            _init_data.video_mode = video_output::mode_from_name(mode_name.toStdString());
+        }
+        else
+        {
+            QString fallback_mode_name = QString(video_output::mode_name(video_output::anaglyph_red_cyan_dubois).c_str());
+            QString mode_name = _settings->value(QString("3d-output-mode"), fallback_mode_name).toString();
+            _init_data.video_mode = video_output::mode_from_name(mode_name.toStdString());
+        }
+        _settings->endGroup();
         _in_out_widget->update(_init_data, true, false);
         _controls_widget->update(_init_data, true, false);
     }
@@ -709,7 +783,9 @@ void main_window::open(QStringList filenames, bool automatic)
 void main_window::file_open()
 {
     QFileDialog *file_dialog = new QFileDialog(this);
-    file_dialog->setDirectory(_last_open_dir);
+    _settings->beginGroup("Session");
+    file_dialog->setDirectory(_settings->value("file-open-dir", QDir::currentPath()).toString());
+    _settings->endGroup();
     file_dialog->setWindowTitle("Open up to three files");
     file_dialog->setAcceptMode(QFileDialog::AcceptOpen);
     file_dialog->setFileMode(QFileDialog::ExistingFiles);
@@ -722,12 +798,14 @@ void main_window::file_open()
     {
         return;
     }
-    _last_open_dir = file_dialog->directory().path();
     if (file_names.size() > 3)
     {
         QMessageBox::critical(this, "Error", "Cannot open more than 3 files");
         return;
     }
+    _settings->beginGroup("Session");
+    _settings->setValue("file-open-dir", file_dialog->directory().path());
+    _settings->endGroup();
     open(file_names, true);
 }
 
@@ -799,6 +877,9 @@ void main_window::help_about()
 player_qt::player_qt() : player(player::slave)
 {
     _qt_app_owner = init_qt();
+    QCoreApplication::setOrganizationName(PACKAGE_NAME);
+    QCoreApplication::setApplicationName(PACKAGE_NAME);
+    _settings = new QSettings;
 }
 
 player_qt::~player_qt()
@@ -807,12 +888,13 @@ player_qt::~player_qt()
     {
         exit_qt();
     }
+    delete _settings;
 }
 
 void player_qt::open(const player_init_data &init_data)
 {
     msg::set_level(init_data.log_level);
-    _main_window = new main_window(init_data);
+    _main_window = new main_window(_settings, init_data);
 }
 
 void player_qt::run()
