@@ -36,7 +36,8 @@
 #include "xgl.h"
 
 
-video_output_opengl::video_output_opengl() throw () : video_output()
+video_output_opengl::video_output_opengl(bool receive_notifications) throw ()
+    : video_output(receive_notifications)
 {
 }
 
@@ -126,6 +127,7 @@ void video_output_opengl::swap_tex_set()
 
 void video_output_opengl::initialize(bool have_pixel_buffer_object, bool have_texture_non_power_of_two, bool have_fragment_shader)
 {
+    _have_valid_data = false;
     _use_non_power_of_two = true;
     if (!have_texture_non_power_of_two)
     {
@@ -139,35 +141,27 @@ void video_output_opengl::initialize(bool have_pixel_buffer_object, bool have_te
     {
         GLint tex_units;
         glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS, &tex_units);
-        if (_mode == anaglyph_red_cyan_monochrome
-                || _mode == anaglyph_red_cyan_full_color
-                || _mode == anaglyph_red_cyan_half_color
-                || _mode == anaglyph_red_cyan_dubois)
+        if (((_mode == even_odd_rows
+                        || _mode == even_odd_columns
+                        || _mode == checkerboard) && tex_units < 7)
+                || ((_mode == anaglyph_red_cyan_monochrome
+                        || _mode == anaglyph_red_cyan_full_color
+                        || _mode == anaglyph_red_cyan_half_color
+                        || _mode == anaglyph_red_cyan_dubois) && tex_units < 6)
+                || tex_units < 3)
         {
-            if (tex_units < 6)
+            _yuv420p_supported = false;
+            if (_src_preferred_frame_format == decoder::frame_format_yuv420p)
             {
-                _yuv420p_supported = false;
-                if (_src_preferred_frame_format == decoder::frame_format_yuv420p)
-                {
-                    msg::wrn("Not enough texture image units for anaglyph output:");
-                    msg::wrn("    Disabling hardware accelerated color space conversion");
-                }
-            }
-        }
-        else
-        {
-            if (tex_units < 3)
-            {
-                _yuv420p_supported = false;
-                if (_src_preferred_frame_format == decoder::frame_format_yuv420p)
-                {
-                    msg::wrn("Not enough texture image units:");
-                    msg::wrn("    Disabling hardware accelerated color space conversion");
-                }
+                msg::wrn("Not enough texture image units for anaglyph output:");
+                msg::wrn("    Disabling hardware accelerated color space conversion");
             }
         }
         std::string mode_str = (
-                _mode == anaglyph_red_cyan_monochrome ? "mode_anaglyph_monochrome"
+                _mode == even_odd_rows ? "mode_even_odd_rows"
+                : _mode == even_odd_columns ? "mode_even_odd_columns"
+                : _mode == checkerboard ? "mode_checkerboard"
+                : _mode == anaglyph_red_cyan_monochrome ? "mode_anaglyph_monochrome"
                 : _mode == anaglyph_red_cyan_full_color ? "mode_anaglyph_full_color"
                 : _mode == anaglyph_red_cyan_half_color ? "mode_anaglyph_half_color"
                 : _mode == anaglyph_red_cyan_dubois ? "mode_anaglyph_dubois"
@@ -178,12 +172,33 @@ void video_output_opengl::initialize(bool have_pixel_buffer_object, bool have_te
         _prg = xgl::CreateProgram("video_output", "", "", fs_src);
         xgl::LinkProgram("video_output", _prg);
         glUseProgram(_prg);
+        if (_mode == even_odd_rows || _mode == even_odd_columns || _mode == checkerboard)
+        {
+            GLubyte even_odd_rows_mask[4] = { 0xff, 0xff, 0x00, 0x00 };
+            GLubyte even_odd_columns_mask[4] = { 0xff, 0x00, 0xff, 0x00 };
+            GLubyte checkerboard_mask[4] = { 0xff, 0x00, 0x00, 0xff };
+            glGenTextures(1, &_mask_tex);
+            glBindTexture(GL_TEXTURE_2D, _mask_tex);
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
+            glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
+            glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
+            glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE8, 2, 2, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE,
+                    _mode == even_odd_rows ? even_odd_rows_mask
+                    : _mode == even_odd_columns ? even_odd_columns_mask
+                    : checkerboard_mask);
+        }
         if (frame_format() == decoder::frame_format_yuv420p)
         {
             glUniform1i(glGetUniformLocation(_prg, "y_l"), 0);
             glUniform1i(glGetUniformLocation(_prg, "u_l"), 1);
             glUniform1i(glGetUniformLocation(_prg, "v_l"), 2);
-            if (_mode == anaglyph_red_cyan_monochrome
+            if (_mode == even_odd_rows
+                    || _mode == even_odd_columns
+                    || _mode == checkerboard
+                    || _mode == anaglyph_red_cyan_monochrome
                     || _mode == anaglyph_red_cyan_full_color
                     || _mode == anaglyph_red_cyan_half_color
                     || _mode == anaglyph_red_cyan_dubois)
@@ -192,17 +207,28 @@ void video_output_opengl::initialize(bool have_pixel_buffer_object, bool have_te
                 glUniform1i(glGetUniformLocation(_prg, "u_r"), 4);
                 glUniform1i(glGetUniformLocation(_prg, "v_r"), 5);
             }
+            if (_mode == even_odd_rows || _mode == even_odd_columns || _mode == checkerboard)
+            {
+                glUniform1i(glGetUniformLocation(_prg, "mask_tex"), 6);
+            }
         }
         else
         {
             glUniform1i(glGetUniformLocation(_prg, "rgb_l"), 0);
-            if (_mode == anaglyph_red_cyan_monochrome
+            if (_mode == even_odd_rows
+                    || _mode == even_odd_columns
+                    || _mode == checkerboard
+                    || _mode == anaglyph_red_cyan_monochrome
                     || _mode == anaglyph_red_cyan_full_color
                     || _mode == anaglyph_red_cyan_half_color
                     || _mode == anaglyph_red_cyan_dubois)
             {
                 glUniform1i(glGetUniformLocation(_prg, "rgb_r"), 1);
             }
+            if (_mode == even_odd_rows || _mode == even_odd_columns || _mode == checkerboard)
+            {
+                glUniform1i(glGetUniformLocation(_prg, "mask_tex"), 2);
+            }
         }
     }
     else
@@ -222,6 +248,10 @@ void video_output_opengl::initialize(bool have_pixel_buffer_object, bool have_te
             msg::wrn("    - Falling back to the low quality anaglyph-full-color method");
             _mode = anaglyph_red_cyan_full_color;
         }
+        else if (_mode == even_odd_rows || _mode == even_odd_columns || _mode == checkerboard)
+        {
+            msg::wrn("    - Falling back to the low quality version of the output method");
+        }
     }
 
     int tex_width = _src_width;
@@ -322,10 +352,15 @@ void video_output_opengl::deinitialize()
     {
         glDeleteTextures(2 * 2, reinterpret_cast<GLuint *>(_rgb_tex));
     }
+    if (_mode == even_odd_rows || _mode == even_odd_columns || _mode == checkerboard)
+    {
+        glDeleteTextures(1, &_mask_tex);
+    }
     if (_pbo != 0)
     {
         glDeleteBuffersARB(1, &_pbo);
     }
+    _have_valid_data = false;
 }
 
 enum decoder::video_frame_format video_output_opengl::frame_format() const
@@ -358,17 +393,17 @@ void video_output_opengl::bind_textures(int unitset, int index)
     }
 }
 
-void video_output_opengl::draw_full_quad()
+void video_output_opengl::draw_quad(float x, float y, float w, float h)
 {
     glBegin(GL_QUADS);
-    glTexCoord2f(0.0f, 0.0f);
-    glVertex2f(-1.0f, 1.0f);
-    glTexCoord2f(_tex_max_x, 0.0f);
-    glVertex2f(1.0f, 1.0f);
-    glTexCoord2f(_tex_max_x, _tex_max_y);
-    glVertex2f(1.0f, -1.0f);
     glTexCoord2f(0.0f, _tex_max_y);
-    glVertex2f(-1.0f, -1.0f);
+    glVertex2f(x, y);
+    glTexCoord2f(_tex_max_x, _tex_max_y);
+    glVertex2f(x + w, y);
+    glTexCoord2f(_tex_max_x, 0.0f);
+    glVertex2f(x + w, y + h);
+    glTexCoord2f(0.0f, 0.0f);
+    glVertex2f(x, y + h);
     glEnd();
 }
 
@@ -458,8 +493,14 @@ static const GLubyte stipple_pattern_checkerboard[132] =
 
 static const GLubyte *stipple_pattern_checkerboard_inv = stipple_pattern_checkerboard + 4;
 
-void video_output_opengl::display(enum video_output::mode mode)
+void video_output_opengl::display(enum video_output::mode mode, float x, float y, float w, float h)
 {
+    glClear(GL_COLOR_BUFFER_BIT);
+    if (!_have_valid_data)
+    {
+        return;
+    }
+
     int64_t display_start = timer::get_microseconds(timer::monotonic);
 
     int left = 0;
@@ -469,7 +510,17 @@ void video_output_opengl::display(enum video_output::mode mode)
         std::swap(left, right);
     }
 
-    glClear(GL_COLOR_BUFFER_BIT);
+    int viewport[4];
+    glGetIntegerv(GL_VIEWPORT, viewport);
+    if ((mode == even_odd_rows || mode == checkerboard) && (window_pos_y() + viewport[1]) % 2 == 0)
+    {
+        std::swap(left, right);
+    }
+    if ((mode == even_odd_columns || mode == checkerboard) && (window_pos_x() + viewport[0]) % 2 == 1)
+    {
+        std::swap(left, right);
+    }
+
     if (_prg != 0)
     {
         glUniform1f(glGetUniformLocation(_prg, "contrast"), _state.contrast);
@@ -477,71 +528,64 @@ void video_output_opengl::display(enum video_output::mode mode)
         glUniform1f(glGetUniformLocation(_prg, "saturation"), _state.saturation);
         glUniform1f(glGetUniformLocation(_prg, "cos_hue"), std::cos(_state.hue * M_PI));
         glUniform1f(glGetUniformLocation(_prg, "sin_hue"), std::sin(_state.hue * M_PI));
+        if (mode == even_odd_rows || mode == even_odd_columns || mode == checkerboard)
+        {
+            glUniform1f(glGetUniformLocation(_prg, "step_x"), _tex_max_x / static_cast<float>(viewport[2]));
+            glUniform1f(glGetUniformLocation(_prg, "step_y"), _tex_max_y / static_cast<float>(viewport[3]));
+        }
     }
 
     if (mode == stereo)
     {
         glDrawBuffer(GL_BACK_LEFT);
         bind_textures(0, left);
-        draw_full_quad();
+        draw_quad(x, y, w, h);
         glDrawBuffer(GL_BACK_RIGHT);
         bind_textures(0, right);
-        draw_full_quad();
+        draw_quad(x, y, w, h);
     }
-    else if (mode == even_odd_rows)
+    else if (mode == even_odd_rows || mode == even_odd_columns || mode == checkerboard)
     {
-        if (window_pos_y() % 2 == 0)
+        if (_prg != 0)
         {
-            std::swap(left, right);
+            float vpw = static_cast<float>(viewport[2]);
+            float vph = static_cast<float>(viewport[3]);
+            bind_textures(0, left);
+            bind_textures(1, right);
+            glActiveTexture(GL_TEXTURE0 + (frame_format() == decoder::frame_format_yuv420p ? 6 : 2));
+            glBindTexture(GL_TEXTURE_2D, _mask_tex);
+            glBegin(GL_QUADS);
+            glTexCoord2f(0.0f, _tex_max_y);
+            glMultiTexCoord2f(GL_TEXTURE1, 0.0f, 0.0f);
+            glVertex2f(x, y);
+            glTexCoord2f(_tex_max_x, _tex_max_y);
+            glMultiTexCoord2f(GL_TEXTURE1, vpw / 2.0f, 0.0f);
+            glVertex2f(x + w, y);
+            glTexCoord2f(_tex_max_x, 0.0f);
+            glMultiTexCoord2f(GL_TEXTURE1, vpw / 2.0f, vph / 2.0f);
+            glVertex2f(x + w, y + h);
+            glTexCoord2f(0.0f, 0.0f);
+            glMultiTexCoord2f(GL_TEXTURE1, 0.0f, vph / 2.0f);
+            glVertex2f(x, y + h);
+            glEnd();
         }
-        glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
-        glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
-        glEnable(GL_POLYGON_STIPPLE);
-        glPolygonStipple(stipple_pattern_even_rows);
-        bind_textures(0, left);
-        draw_full_quad();
-        glPolygonStipple(stipple_pattern_odd_rows);
-        bind_textures(0, right);
-        draw_full_quad();
-        glDisable(GL_POLYGON_STIPPLE);
-    }
-    else if (mode == even_odd_columns)
-    {
-        if (window_pos_x() % 2 == 1)
+        else
         {
-            std::swap(left, right);
+            const GLubyte *left_stipple = (mode == even_odd_rows ? stipple_pattern_even_rows
+                    : mode == even_odd_columns ? stipple_pattern_even_cols : stipple_pattern_checkerboard);
+            const GLubyte *right_stipple = (mode == even_odd_rows ? stipple_pattern_odd_rows
+                    : mode == even_odd_columns ? stipple_pattern_odd_cols : stipple_pattern_checkerboard_inv);
+            glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
+            glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
+            glEnable(GL_POLYGON_STIPPLE);
+            glPolygonStipple(left_stipple);
+            bind_textures(0, left);
+            draw_quad(x, y, w, h);
+            glPolygonStipple(right_stipple);
+            bind_textures(0, right);
+            draw_quad(x, y, w, h);
+            glDisable(GL_POLYGON_STIPPLE);
         }
-        glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
-        glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
-        glEnable(GL_POLYGON_STIPPLE);
-        glPolygonStipple(stipple_pattern_even_cols);
-        bind_textures(0, left);
-        draw_full_quad();
-        glPolygonStipple(stipple_pattern_odd_cols);
-        bind_textures(0, right);
-        draw_full_quad();
-        glDisable(GL_POLYGON_STIPPLE);
-    }
-    else if (mode == checkerboard)
-    {
-        if (window_pos_y() % 2 == 0)
-        {
-            std::swap(left, right);
-        }
-        if (window_pos_x() % 2 == 1)
-        {
-            std::swap(left, right);
-        }
-        glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
-        glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
-        glEnable(GL_POLYGON_STIPPLE);
-        glPolygonStipple(stipple_pattern_checkerboard);
-        bind_textures(0, left);
-        draw_full_quad();
-        glPolygonStipple(stipple_pattern_checkerboard_inv);
-        bind_textures(0, right);
-        draw_full_quad();
-        glDisable(GL_POLYGON_STIPPLE);
     }
     else if (_prg != 0
             && (mode == anaglyph_red_cyan_monochrome
@@ -551,27 +595,27 @@ void video_output_opengl::display(enum video_output::mode mode)
     {
         bind_textures(0, left);
         bind_textures(1, right);
-        draw_full_quad();
+        draw_quad(x, y, w, h);
     }
     else if (_prg == 0 && mode == anaglyph_red_cyan_full_color)
     {
         glColorMask(GL_TRUE, GL_FALSE, GL_FALSE, GL_FALSE);
         bind_textures(0, left);
-        draw_full_quad();
+        draw_quad(x, y, w, h);
         glColorMask(GL_FALSE, GL_TRUE, GL_TRUE, GL_FALSE);
         bind_textures(0, right);
-        draw_full_quad();
+        draw_quad(x, y, w, h);
         glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
     }
     else if (mode == mono_left)
     {
         bind_textures(0, left);
-        draw_full_quad();
+        draw_quad(x, y, w, h);
     }
     else if (mode == mono_right)
     {
         bind_textures(0, right);
-        draw_full_quad();
+        draw_quad(x, y, w, h);
     }
     else if (mode == left_right || mode == left_right_half)
     {
@@ -715,6 +759,12 @@ void video_output_opengl::prepare(
         uint8_t *l_data[3], size_t l_line_size[3],
         uint8_t *r_data[3], size_t r_line_size[3])
 {
+    if (!l_data[0])
+    {
+        _have_valid_data = false;
+        return;
+    }
+
     int64_t prepare_start = timer::get_microseconds(timer::monotonic);
 
     int tex_set = (_active_tex_set == 0 ? 1 : 0);
@@ -757,6 +807,7 @@ void video_output_opengl::prepare(
                     GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, r_data[0]);
         }
     }
+    _have_valid_data = true;
 
     int64_t prepare_stop = timer::get_microseconds(timer::monotonic);
     msg::dbg("texture upload: %g seconds", (prepare_stop - prepare_start) / 1e6f);
